@@ -1,5 +1,5 @@
 import os
 import glob
 __all__ = [os.path.basename(f)[:-3]
-           for f in glob.glob(os.path.dirname(__file__) + "/*.py")
-           if not os.path.basename(f).startswith('_')]
+           for f in glob.glob(os.path.join(os.path.dirname(__file__),"*.py"))
+           if not os.path.basename(f).startswith('_')]
\ No newline at end of file
